@@ -250,40 +250,43 @@ static int get_range(struct supplicant_task *task)
 	return err;
 }
 
-static char *get_bssid(struct connman_device *device)
+static int get_bssid(struct connman_device *device,
+				unsigned char *bssid, unsigned int *bssid_len)
 {
-	char *bssid;
-	unsigned char ioctl_bssid[ETH_ALEN];
-	int fd, ret;
 	struct iwreq wrq;
+	char *ifname;
+	int ifindex;
+	int fd, err;
 
-	if (connman_device_get_type(device) != CONNMAN_DEVICE_TYPE_WIFI)
-		return NULL;
+	ifindex = connman_device_get_index(device);
+	if (ifindex < 0)
+		return -EINVAL;
+
+	ifname = connman_inet_ifname(ifindex);
+	if (ifname == NULL)
+		return -EINVAL;
 
 	fd = socket(PF_INET, SOCK_DGRAM, 0);
-	if (fd < 0)
-		return NULL;
+	if (fd < 0) {
+		g_free(ifname);
+		return -EINVAL;
+	}
 
 	memset(&wrq, 0, sizeof(wrq));
-	strncpy(wrq.ifr_name, connman_device_get_interface(device), IFNAMSIZ);
+	strncpy(wrq.ifr_name, ifname, IFNAMSIZ);
 
-	ret = ioctl(fd, SIOCGIWAP, &wrq);
+	err = ioctl(fd, SIOCGIWAP, &wrq);
+
+	g_free(ifname);
 	close(fd);
-	if (ret != 0)
-		return NULL;
 
-	memcpy(ioctl_bssid, wrq.u.ap_addr.sa_data, ETH_ALEN);
+	if (err < 0)
+		return -EIO;
 
-	bssid = g_try_malloc0(13);
-	if (bssid == NULL)
-		return NULL;
+	memcpy(bssid, wrq.u.ap_addr.sa_data, ETH_ALEN);
+	*bssid_len = ETH_ALEN;
 
-	snprintf(bssid, 13, "%02x%02x%02x%02x%02x%02x",
-		 ioctl_bssid[0], ioctl_bssid[1],
-		 ioctl_bssid[2], ioctl_bssid[3],
-		 ioctl_bssid[4], ioctl_bssid[5]);
-
-	return bssid;
+	return 0;
 }
 
 static void add_interface_reply(DBusPendingCall *call, void *user_data)
@@ -485,10 +488,12 @@ static int remove_interface(struct supplicant_task *task)
 
 	DBG("task %p", task);
 
+#if 0
 	if (task->created == FALSE) {
 		connman_device_set_powered(task->device, FALSE);
 		return 0;
 	}
+#endif
 
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, SUPPLICANT_PATH,
 					SUPPLICANT_INTF, "removeInterface");
@@ -936,6 +941,11 @@ static char *build_group(const char *addr, const char *name,
 	if (str == NULL)
 		return NULL;
 
+	if (ssid == NULL) {
+		g_string_append_printf(str, "hidden_%s", addr);
+		goto done;
+	}
+
 	for (i = 0; special_ssid[i].name; i++) {
 		if (g_strcmp0(special_ssid[i].name, name) == 0) {
 			if (special_ssid[i].value == NULL)
@@ -964,7 +974,7 @@ static void extract_addr(DBusMessageIter *value,
 					struct supplicant_result *result)
 {
 	DBusMessageIter array;
-	struct ether_addr *eth;
+	struct ether_addr eth;
 	unsigned char *addr;
 	int addr_len;
 
@@ -985,15 +995,14 @@ static void extract_addr(DBusMessageIter *value,
 	if (result->path == NULL)
 		return;
 
-	eth = (void *) addr;
-
+	memcpy(&eth, addr, sizeof(eth));
 	snprintf(result->path, 13, "%02x%02x%02x%02x%02x%02x",
-						eth->ether_addr_octet[0],
-						eth->ether_addr_octet[1],
-						eth->ether_addr_octet[2],
-						eth->ether_addr_octet[3],
-						eth->ether_addr_octet[4],
-						eth->ether_addr_octet[5]);
+						eth.ether_addr_octet[0],
+						eth.ether_addr_octet[1],
+						eth.ether_addr_octet[2],
+						eth.ether_addr_octet[3],
+						eth.ether_addr_octet[4],
+						eth.ether_addr_octet[5]);
 }
 
 static void extract_ssid(DBusMessageIter *value,
@@ -1009,6 +1018,9 @@ static void extract_ssid(DBusMessageIter *value,
 	if (ssid_len < 1)
 		return;
 
+	if (ssid[0] == '\0')
+		return;
+
 	result->ssid = g_try_malloc(ssid_len);
 	if (result->ssid == NULL)
 		return;
@@ -1124,7 +1136,7 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	unsigned char strength;
 	unsigned short channel, frequency;
 	const char *mode, *security;
-	char *group;
+	char *group = NULL;
 	unsigned int ssid_len;
 
 	DBG("task %p", task);
@@ -1281,11 +1293,12 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	connman_network_set_uint16(network, "WiFi.Channel", channel);
 	connman_network_set_string(network, "WiFi.Security", security);
 
-	connman_network_set_group(network, group);
-
-	g_free(group);
+	if (result.ssid != NULL)
+		connman_network_set_group(network, group);
 
 done:
+	g_free(group);
+
 	g_free(result.path);
 	g_free(result.addr);
 	g_free(result.name);
@@ -1336,8 +1349,7 @@ static void get_properties(struct supplicant_task *task)
 	return;
 
 noscan:
-	if (task->noscan == FALSE)
-		connman_device_set_scanning(task->device, FALSE);
+	connman_device_set_scanning(task->device, FALSE);
 }
 
 static void scan_results_reply(DBusPendingCall *call, void *user_data)
@@ -1394,8 +1406,7 @@ done:
 	dbus_message_unref(reply);
 
 noscan:
-	if (task->noscan == FALSE)
-		connman_device_set_scanning(task->device, FALSE);
+	connman_device_set_scanning(task->device, FALSE);
 }
 
 static void scan_results_available(struct supplicant_task *task)
@@ -1483,13 +1494,15 @@ static int task_connect(struct supplicant_task *task)
 
 	enable_network(task);
 
-	return 0;
+	return -EINPROGRESS;
 }
 
 static void state_change(struct supplicant_task *task, DBusMessage *msg)
 {
 	DBusError error;
 	const char *newstate, *oldstate;
+	unsigned char bssid[ETH_ALEN];
+	unsigned int bssid_len;
 	enum supplicant_state state;
 
 	dbus_error_init(&error);
@@ -1541,32 +1554,9 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 
 	switch (task->state) {
 	case WPA_COMPLETED:
-		if (connman_network_get_group(task->network) == NULL) {
-			const char *name, *mode, *security;
-			char *group, *bssid;
-
-			/*
-			 * This is a hidden network, we need to set its
-			 * group based on the BSSID we just joined.
-			 */
-			bssid = get_bssid(task->device);
-
-			name = connman_network_get_string(task->network,
-								"Name");
-			mode = connman_network_get_string(task->network,
-								"WiFi.Mode");
-			security = connman_network_get_string(task->network,
-							"WiFi.Security");
-
-			if (bssid && name && mode && security) {
-				group = build_group(bssid, name, NULL, 0,
-								mode, security);
-				connman_network_set_group(task->network, group);
-			}
-
-			g_free(bssid);
-			g_free(group);
-		}
+		if (get_bssid(task->device, bssid, &bssid_len) == 0)
+	                connman_network_set_address(task->network,
+							bssid, bssid_len);
 
 		/* carrier on */
 		connman_network_set_connected(task->network, TRUE);
@@ -1574,6 +1564,8 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 		break;
 
 	case WPA_DISCONNECTED:
+		disable_network(task);
+
 		if (task->disconnecting == TRUE) {
 			connman_network_set_connected(task->network, FALSE);
 			connman_network_unref(task->network);
@@ -1595,6 +1587,22 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 		connman_network_set_associating(task->network, TRUE);
 		break;
 
+	case WPA_INACTIVE:
+		remove_network(task);
+
+		if (task->disconnecting == TRUE) {
+			connman_network_set_connected(task->network, FALSE);
+			connman_network_unref(task->network);
+			task->disconnecting = FALSE;
+
+			if (task->pending_network != NULL) {
+				task->network = task->pending_network;
+				task->pending_network = NULL;
+				task_connect(task);
+			}
+		}
+		break;
+
 	default:
 		connman_network_set_associating(task->network, FALSE);
 		break;
@@ -1753,7 +1761,7 @@ int supplicant_connect(struct connman_network *network)
 		return task_connect(task);
 	}
 
-	return 0;
+	return -EINPROGRESS;
 }
 
 int supplicant_disconnect(struct connman_network *network)
