@@ -187,6 +187,8 @@ static const char *error2string(enum connman_service_error error)
 	switch (error) {
 	case CONNMAN_SERVICE_ERROR_UNKNOWN:
 		break;
+	case CONNMAN_SERVICE_ERROR_OUT_OF_RANGE:
+		return "out-of-range";
 	case CONNMAN_SERVICE_ERROR_PIN_MISSING:
 		return "pin-missing";
 	case CONNMAN_SERVICE_ERROR_DHCP_FAILED:
@@ -394,6 +396,13 @@ static DBusMessage *set_property(DBusConnection *conn,
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
+static void set_idle(struct connman_service *service)
+{
+	service->state = CONNMAN_SERVICE_STATE_IDLE;
+	service->error = CONNMAN_SERVICE_ERROR_UNKNOWN;
+	state_changed(service);
+}
+
 static DBusMessage *clear_property(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
@@ -410,9 +419,7 @@ static DBusMessage *clear_property(DBusConnection *conn,
 		return __connman_error_permission_denied(msg);
 
 	if (g_str_equal(name, "Error") == TRUE) {
-		service->state = CONNMAN_SERVICE_STATE_IDLE;
-		service->error = CONNMAN_SERVICE_ERROR_UNKNOWN;
-		state_changed(service);
+		set_idle(service);
 
 		g_get_current_time(&service->modified);
 		__connman_storage_save_service(service);
@@ -440,24 +447,6 @@ static connman_bool_t is_connecting(struct connman_service *service)
 	return FALSE;
 }
 
-static struct connman_service *find_pending_service(void)
-{
-	struct connman_service *service;
-	GSequenceIter *iter;
-
-	iter = g_sequence_get_begin_iter(service_list);
-
-	while (g_sequence_iter_is_end(iter) == FALSE) {
-		service = g_sequence_get(iter);
-		if (service->pending != NULL)
-			return service;
-
-		iter = g_sequence_iter_next(iter);
-	}
-
-	return NULL;
-}
-
 static connman_bool_t is_ignore(struct connman_service *service)
 {
 	if (service->ignore == TRUE)
@@ -471,41 +460,65 @@ static connman_bool_t is_ignore(struct connman_service *service)
 
 void __connman_service_auto_connect(void)
 {
-	struct connman_service *service;
+	struct connman_service *service = NULL;
 	GSequenceIter *iter;
 
 	DBG("");
 
-	service = find_pending_service();
-	if (service != NULL)
-		return;
-
 	iter = g_sequence_get_begin_iter(service_list);
-	if (g_sequence_iter_is_end(iter) == TRUE)
-		return;
 
-	service = g_sequence_get(iter);
-
-	while (is_ignore(service) == TRUE) {
-		iter = g_sequence_iter_next(iter);
-		if (g_sequence_iter_is_end(iter))
-			return;
+	while (g_sequence_iter_is_end(iter) == FALSE) {
 		service = g_sequence_get(iter);
+
+		if (service->pending != NULL)
+			return;
+
+		if (is_connecting(service) == TRUE)
+			return;
+
+		if (service->favorite == FALSE)
+			return;
+
+		if (service->state == CONNMAN_SERVICE_STATE_READY)
+			return;
+
+		if (is_ignore(service) == FALSE &&
+				service->state == CONNMAN_SERVICE_STATE_IDLE)
+			break;
+
+		service = NULL;
+
+		iter = g_sequence_iter_next(iter);
 	}
 
-	if (service->favorite == FALSE)
-		return;
-
-	if (service->state == CONNMAN_SERVICE_STATE_READY)
-		return;
-
-	if (is_connecting(service) == TRUE)
-		return;
-
-	if (service->state == CONNMAN_SERVICE_STATE_IDLE)
+	if (service != NULL)
 		__connman_service_connect(service);
 }
 
+static void reply_pending(struct connman_service *service, int error)
+{
+	if (service->timeout > 0) {
+		g_source_remove(service->timeout);
+		service->timeout = 0;
+	}
+
+	if (service->pending != NULL) {
+		if (error > 0) {
+			DBusMessage *reply;
+
+			reply = __connman_error_failed(service->pending,
+								error);
+			if (reply != NULL)
+				g_dbus_send_message(connection, reply);
+		} else
+			g_dbus_send_reply(connection, service->pending,
+							DBUS_TYPE_INVALID);
+
+		dbus_message_unref(service->pending);
+		service->pending = NULL;
+	}
+}
+
 static gboolean connect_timeout(gpointer user_data)
 {
 	struct connman_service *service = user_data;
@@ -564,6 +577,7 @@ static DBusMessage *connect_service(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
 	struct connman_service *service = user_data;
+	GSequenceIter *iter;
 	int err;
 
 	DBG("service %p", service);
@@ -571,6 +585,17 @@ static DBusMessage *connect_service(DBusConnection *conn,
 	if (service->pending != NULL)
 		return __connman_error_in_progress(msg);
 
+	iter = g_sequence_get_begin_iter(service_list);
+
+	while (g_sequence_iter_is_end(iter) == FALSE) {
+		struct connman_service *service = g_sequence_get(iter);
+
+		if (is_connecting(service) == TRUE)
+			return __connman_error_in_progress(msg);
+
+		iter = g_sequence_iter_next(iter);
+	}
+
 	service->ignore = FALSE;
 
 	service->pending = dbus_message_ref(msg);
@@ -601,18 +626,8 @@ static DBusMessage *disconnect_service(DBusConnection *conn,
 
 	DBG("service %p", service);
 
-	if (service->pending != NULL) {
-		DBusMessage *reply;
-
-		reply = __connman_error_operation_aborted(service->pending);
-		if (reply != NULL)
-			g_dbus_send_message(conn, reply);
-
-		dbus_message_unref(service->pending);
-		service->pending = NULL;
-
-		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
-	}
+	if (service->pending != NULL)
+		reply_pending(service, ECONNABORTED);
 
 	service->ignore = TRUE;
 
@@ -741,20 +756,12 @@ static void service_free(gpointer user_data)
 
 	g_hash_table_remove(service_hash, service->identifier);
 
-	if (service->timeout > 0) {
-		g_source_remove(service->timeout);
-		service->timeout = 0;
-	}
-
-	if (service->pending != NULL) {
-		dbus_message_unref(service->pending);
-		service->pending = NULL;
-	}
+	reply_pending(service, ENOENT);
 
 	service->path = NULL;
 
 	if (path != NULL) {
-		__connman_profile_changed(TRUE);
+		__connman_profile_changed(FALSE);
 
 		g_dbus_unregister_interface(connection, path,
 						CONNMAN_SERVICE_INTERFACE);
@@ -787,9 +794,17 @@ void __connman_service_put(struct connman_service *service)
 		GSequenceIter *iter;
 
 		iter = g_hash_table_lookup(service_hash, service->identifier);
-		if (iter != NULL)
+		if (iter != NULL) {
+			reply_pending(service, ENOENT);
+
+			__connman_service_disconnect(service);
+
+			service->state = CONNMAN_SERVICE_STATE_FAILURE;
+			service->error = CONNMAN_SERVICE_ERROR_OUT_OF_RANGE;
+			state_changed(service);
+
 			g_sequence_remove(iter);
-		else
+		} else
 			service_free(service);
 	}
 }
@@ -985,6 +1000,8 @@ int __connman_service_indicate_state(struct connman_service *service,
 			service->state != CONNMAN_SERVICE_STATE_DISCONNECT) {
 		service->state = CONNMAN_SERVICE_STATE_DISCONNECT;
 		state_changed(service);
+
+		__connman_service_disconnect(service);
 	}
 
 	service->state = state;
@@ -993,39 +1010,14 @@ int __connman_service_indicate_state(struct connman_service *service,
 	if (state == CONNMAN_SERVICE_STATE_READY) {
 		connman_service_set_favorite(service, TRUE);
 
-		if (service->timeout > 0) {
-			g_source_remove(service->timeout);
-			service->timeout = 0;
-		}
-
-		if (service->pending != NULL) {
-			g_dbus_send_reply(connection, service->pending,
-							DBUS_TYPE_INVALID);
-
-			dbus_message_unref(service->pending);
-			service->pending = NULL;
-		}
+		reply_pending(service, 0);
 
 		g_get_current_time(&service->modified);
 		__connman_storage_save_service(service);
 	}
 
 	if (state == CONNMAN_SERVICE_STATE_FAILURE) {
-		if (service->timeout > 0) {
-			g_source_remove(service->timeout);
-			service->timeout = 0;
-		}
-
-		if (service->pending != NULL) {
-			DBusMessage *reply;
-
-			reply = __connman_error_failed(service->pending, EIO);
-			if (reply != NULL)
-				g_dbus_send_message(connection, reply);
-
-			dbus_message_unref(service->pending);
-			service->pending = NULL;
-		}
+		reply_pending(service, EIO);
 
 		g_get_current_time(&service->modified);
 		__connman_storage_save_service(service);
@@ -1038,6 +1030,10 @@ int __connman_service_indicate_state(struct connman_service *service,
 
 	__connman_profile_changed(FALSE);
 
+	if (service->state == CONNMAN_SERVICE_STATE_IDLE ||
+			service->state == CONNMAN_SERVICE_STATE_FAILURE)
+		__connman_element_request_scan(CONNMAN_ELEMENT_TYPE_UNKNOWN);
+
 	return 0;
 }
 
@@ -1148,15 +1144,72 @@ static struct connman_service *__connman_service_lookup(const char *identifier)
 	return NULL;
 }
 
+static struct connman_network *create_hidden_wifi(struct connman_device *device,
+		const char *ssid, const char *mode, const char *security)
+{
+	struct connman_network *network;
+	char *name;
+	int index;
+	unsigned int i, ssid_len;
+
+	ssid_len = strlen(ssid);
+	if (ssid_len < 1)
+		return NULL;
+
+	network = connman_network_create(NULL, CONNMAN_NETWORK_TYPE_WIFI);
+	if (network == NULL)
+		return NULL;
+
+	connman_network_set_blob(network, "WiFi.SSID",
+					(unsigned char *) ssid, ssid_len);
+
+	connman_network_set_string(network, "WiFi.Mode", mode);
+	connman_network_set_string(network, "WiFi.Security", security);
+
+	name = g_try_malloc0(ssid_len + 1);
+	if (name == NULL) {
+		connman_network_unref(network);
+		return NULL;
+	}
+
+	for (i = 0; i < ssid_len; i++) {
+		if (g_ascii_isprint(ssid[i]))
+			name[i] = ssid[i];
+		else
+			name[i] = ' ';
+	}
+
+	connman_network_set_name(network, name);
+
+	g_free(name);
+
+	index = connman_device_get_index(device);
+	connman_network_set_index(network, index);
+
+	connman_network_set_protocol(network, CONNMAN_NETWORK_PROTOCOL_IP);
+
+	if (connman_device_add_network(device, network) < 0) {
+		connman_network_unref(network);
+		return NULL;
+	}
+
+	connman_network_set_available(network, TRUE);
+
+	return network;
+}
+
 int __connman_service_create_and_connect(DBusMessage *msg)
 {
 	struct connman_service *service;
+	struct connman_network *network;
 	struct connman_device *device;
 	DBusMessageIter iter, array;
 	const char *mode = "managed", *security = "none";
 	const char *type = NULL, *ssid = NULL, *passphrase = NULL;
+	unsigned int ssid_len = 0;
 	const char *ident;
 	char *name, *group;
+	gboolean created = FALSE;
 	int err;
 
 	dbus_message_iter_init(msg, &iter);
@@ -1202,6 +1255,10 @@ int __connman_service_create_and_connect(DBusMessage *msg)
 	if (ssid == NULL)
 		return -EINVAL;
 
+	ssid_len = strlen(ssid);
+	if (ssid_len < 1)
+		return -EINVAL;
+
 	device = __connman_element_find_device(CONNMAN_DEVICE_TYPE_WIFI);
 	if (device == NULL)
 		return -EOPNOTSUPP;
@@ -1211,36 +1268,64 @@ int __connman_service_create_and_connect(DBusMessage *msg)
 		return -EOPNOTSUPP;
 
 	group = connman_wifi_build_group_name((unsigned char *) ssid,
-						strlen(ssid), mode, security);
+						ssid_len, mode, security);
 	if (group == NULL)
 		return -EINVAL;
 
 	name = g_strdup_printf("%s_%s_%s", type, ident, group);
 
-	g_free(group);
+	service = __connman_service_lookup(name);
+
+	if (service != NULL)
+		goto done;
+
+	network = create_hidden_wifi(device, ssid, mode, security);
+	if (network != NULL) {
+		connman_network_set_group(network, group);
+		created = TRUE;
+	}
 
 	service = __connman_service_lookup(name);
 
+done:
 	g_free(name);
+	g_free(group);
 
-	if (service != NULL) {
-		if (passphrase != NULL) {
-			g_free(service->passphrase);
-			service->passphrase = g_strdup(passphrase);
-		}
-
-		err = __connman_service_connect(service);
-		if (err < 0 && err != -EINPROGRESS)
-			return err;
-
-		g_dbus_send_reply(connection, msg,
-					DBUS_TYPE_OBJECT_PATH, &service->path,
-							DBUS_TYPE_INVALID);
-
-		return 0;
+	if (service == NULL) {
+		err = -EOPNOTSUPP;
+		goto failed;
 	}
 
-	return -EOPNOTSUPP;
+	__connman_device_disconnect(device);
+
+	if (passphrase != NULL) {
+		g_free(service->passphrase);
+		service->passphrase = g_strdup(passphrase);
+	}
+
+	err = __connman_service_connect(service);
+	if (err < 0 && err != -EINPROGRESS)
+		goto failed;
+
+	g_dbus_send_reply(connection, msg,
+				DBUS_TYPE_OBJECT_PATH, &service->path,
+							DBUS_TYPE_INVALID);
+
+	return 0;
+
+failed:
+	if (service != NULL && created == TRUE) {
+		struct connman_network *network = service->network;
+
+		if (network != NULL) {
+			connman_network_set_available(network, FALSE);
+			__connman_device_cleanup_networks(device);
+		}
+
+		__connman_service_put(service);
+	}
+
+	return err;
 }
 
 /**
@@ -1313,7 +1398,7 @@ static int service_register(struct connman_service *service)
 }
 
 /**
- * connman_service_lookup_from_device:
+ * __connman_service_lookup_from_device:
  * @device: device structure
  *
  * Look up a service by device (reference count will not be increased)
@@ -1363,7 +1448,7 @@ static enum connman_service_type convert_device_type(struct connman_device *devi
 }
 
 /**
- * connman_service_create_from_device:
+ * __connman_service_create_from_device:
  * @device: device structure
  *
  * Look up service by device and if not found, create one
@@ -1406,8 +1491,19 @@ done:
 	return service;
 }
 
+void __connman_service_remove_from_device(struct connman_device *device)
+{
+	struct connman_service *service;
+
+	service = __connman_service_lookup_from_device(device);
+	if (service == NULL)
+		return;
+
+	__connman_service_put(service);
+}
+
 /**
- * connman_service_lookup_from_network:
+ * __connman_service_lookup_from_network:
  * @network: network structure
  *
  * Look up a service by network (reference count will not be increased)
@@ -1443,15 +1539,20 @@ unsigned int __connman_service_get_order(struct connman_service *service)
 	if (service == NULL)
 		return 0;
 
-	if (service->favorite == FALSE)
-		return 0;
+	if (service->favorite == FALSE) {
+		service->order = 0;
+		goto done;
+	}
 
 	iter = g_hash_table_lookup(service_hash, service->identifier);
 	if (iter != NULL) {
 		if (g_sequence_iter_get_position(iter) == 0)
-			return 1;
+			service->order = 1;
+		else
+			service->order = 0;
 	}
 
+done:
 	return service->order;
 }
 
@@ -1539,28 +1640,21 @@ static void update_from_network(struct connman_service *service,
 
 	if (service->strength > strength && service->network != NULL) {
 		connman_network_unref(service->network);
-		service->network = NULL;
+		service->network = connman_network_ref(network);
 
 		strength_changed(service);
 	}
 
-	if (service->network == NULL) {
+	if (service->network == NULL)
 		service->network = connman_network_ref(network);
 
-		str = connman_network_get_string(network, "WiFi.Passphrase");
-		if (str != NULL) {
-			g_free(service->passphrase);
-			service->passphrase = g_strdup(str);
-		}
-	}
-
 	iter = g_hash_table_lookup(service_hash, service->identifier);
 	if (iter != NULL)
 		g_sequence_sort_changed(iter, service_compare, NULL);
 }
 
 /**
- * connman_service_create_from_network:
+ * __connman_service_create_from_network:
  * @network: network structure
  *
  * Look up service by network and if not found, create one
@@ -1573,6 +1667,17 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 
 	service = __connman_service_lookup_from_network(network);
 	if (service != NULL) {
+		if (g_atomic_int_get(&service->refcount) == 0) {
+			if (service->timeout > 0) {
+				g_source_remove(service->timeout);
+				service->timeout = 0;
+			}
+
+			set_idle(service);
+		}
+
+		connman_service_ref(service);
+
 		update_from_network(service, network);
 		return service;
 	}
@@ -1585,6 +1690,9 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 	if (group == NULL)
 		return NULL;
 
+	if (__connman_network_get_weakness(network) == TRUE)
+		return NULL;
+
 	name = g_strdup_printf("%s_%s_%s",
 			__connman_network_get_type(network), ident, group);
 
@@ -1619,6 +1727,17 @@ done:
 	return service;
 }
 
+void __connman_service_remove_from_network(struct connman_network *network)
+{
+	struct connman_service *service;
+
+	service = __connman_service_lookup_from_network(network);
+	if (service == NULL)
+		return;
+
+	__connman_service_put(service);
+}
+
 static int service_load(struct connman_service *service)
 {
 	GKeyFile *keyfile;
@@ -1842,7 +1961,7 @@ update:
 		g_free(str);
 	}
 
-	if (service->passphrase != NULL)
+	if (service->passphrase != NULL && strlen(service->passphrase) > 0)
 		g_key_file_set_string(keyfile, service->identifier,
 					"Passphrase", service->passphrase);
 	else
